/* Rotary encoder handler for arduino. v1.1
 *
 * Copyright 2011 Ben Buxton. Licenced under the GNU GPL Version 3.
 * Contact: bb@cactii.net
 *
 * A typical mechanical rotary encoder emits a two bit gray code
 * on 3 output pins. Every step in the output (often accompanied
 * by a physical 'click') generates a specific sequence of output
 * codes on the pins.
 *
 * There are 3 pins used for the rotary encoding - one common and
 * two 'bit' pins.
 *
 * The following is the typical sequence of code on the output when
 * moving from one step to the next:
 *
 *   Position   Bit1   Bit2
 *   ----------------------
 *     Step1     0      0
 *      1/4      1      0
 *      1/2      1      1
 *      3/4      0      1
 *     Step2     0      0
 *
 * From this table, we can see that when moving from one 'click' to
 * the next, there are 4 changes in the output code.
 *
 * - From an initial 0 - 0, Bit1 goes high, Bit0 stays low.
 * - Then both bits are high, halfway through the step.
 * - Then Bit1 goes low, but Bit2 stays high.
 * - Finally at the end of the step, both bits return to 0.
 *
 * Detecting the direction is easy - the table simply goes in the other
 * direction (read up instead of down).
 *
 * To decode this, we use a simple state machine. Every time the output
 * code changes, it follows state, until finally a full steps worth of
 * code is received (in the correct order). At the final 0-0, it returns
 * a value indicating a step in one direction or the other.
 *
 * It's also possible to use 'half-step' mode. This just emits an event
 * at both the 0-0 and 1-1 positions. This might be useful for some
 * encoders where you want to detect all positions.
 *
 * If an invalid state happens (for example we go from '0-1' straight
 * to '1-0'), the state machine resets to the start until 0-0 and the
 * next valid codes occur.
 *
 * The biggest advantage of using a state machine over other algorithms
 * is that this has inherent debounce built in. Other algorithms emit spurious
 * output with switch bounce, but this one will simply flip between
 * sub-states until the bounce settles, then continue along the state
 * machine.
 * A side effect of debounce is that fast rotations can cause steps to
 * be skipped. By not requiring debounce, fast rotations can be accurately
 * measured.
 * Another advantage is the ability to properly handle bad state, such
 * as due to EMI, etc.
 * It is also a lot simpler than others - a static state table and less
 * than 10 lines of logic.
 */

#include "sdkconfig.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "esp_attr.h"
#include "encoders.h"
#include "driver/gpio.h"


#include <stdio.h>
#include <string.h>



/*
 * The below state table has, for each state (row), the new state
 * to set based on the next encoder output. From left to right in,
 * the table, the encoder outputs are 00, 01, 10, 11, and the value
 * in that position is the new state to set.
 */

// Values returned by 'process'
// No complete step yet.
#define DIR_NONE 0x0
// Clockwise step.
#define DIR_CW 0x10
// Anti-clockwise step.
#define DIR_CCW 0x20
#define R_START 0x0

// Use the full-step state table (emits a code at 00 only)
#define R_CW_FINAL 0x1
#define R_CW_BEGIN 0x2
#define R_CW_NEXT 0x3
#define R_CCW_BEGIN 0x4
#define R_CCW_FINAL 0x5
#define R_CCW_NEXT 0x6

#define GPIO_ALL_IN  0b1111111100001110111111111111111111111111ULL

const uint8_t ttable[7][4] = {
  // R_START
  {R_START,    R_CW_BEGIN,  R_CCW_BEGIN, R_START},
  // R_CW_FINAL
  {R_CW_NEXT,  R_START,     R_CW_FINAL,  R_START | DIR_CW},
  // R_CW_BEGIN
  {R_CW_NEXT,  R_CW_BEGIN,  R_START,     R_START},
  // R_CW_NEXT
  {R_CW_NEXT,  R_CW_BEGIN,  R_CW_FINAL,  R_START},
  // R_CCW_BEGIN
  {R_CCW_NEXT, R_START,     R_CCW_BEGIN, R_START},
  // R_CCW_FINAL
  {R_CCW_NEXT, R_CCW_FINAL, R_START,     R_START | DIR_CCW},
  // R_CCW_NEXT
  {R_CCW_NEXT, R_CCW_FINAL, R_CCW_BEGIN, R_START},
};

static xQueueHandle queue = NULL;
static TaskHandle_t task = NULL;
static uint8_t attached = 0;

/*
 * Helper functions
 */

static void encoder_task(void *arg) {
	encoder_deferred_data_t data;

    for(;;) {
        xQueueReceive(queue, &data, portMAX_DELAY);

        if (data.h->callback) {
        	data.h->callback(data.h->callback_id, data.dir, data.counter);
        }
    }
}

static void IRAM_ATTR encoder_isr(void* arg) {

	encoder_h_t *encoder = (encoder_h_t *)arg;
	encoder_deferred_data_t data;
	uint8_t result, has_data;
	int8_t dir;

	has_data = 0;
	dir = 0;

	uint8_t A = (uint8_t) gpio_get_level(encoder->A);
	uint8_t B = (uint8_t) gpio_get_level(encoder->B);


	// Grab state of input pins.
	uint8_t pinstate = (A << 1) | B;

	// Determine new state from the pins and state table.
	encoder->state = ttable[encoder->state & 0xf][pinstate];
	result = encoder->state & 0x30;

	if (result == DIR_CW) {
		encoder->counter++;
		has_data = 1;
		dir = 1;
	} else if (result == DIR_CCW) {
		encoder->counter--;
		has_data = 1;
		dir = -1;
	}

	if (has_data && encoder->callback) {
		if (encoder->deferred) {
			data.h = encoder;
			data.counter = encoder->counter;
			data.dir = dir;
			xQueueSendFromISR(queue, &data, NULL);
		} else {
			encoder->callback(encoder->callback_id, dir, encoder->counter);
		}
	}
}

/*
 * Operation functions
 */

esp_err_t encoder_setup(int8_t a, int8_t b, encoder_h_t **h) {
	// Sanity checks
	if ((!GPIO_IS_VALID_GPIO(a)) && ((GPIO_ALL_IN & (1ULL << a)))) {
		return ESP_ERR_INVALID_ARG;
	}

	if ((!GPIO_IS_VALID_GPIO(b)) && ((GPIO_ALL_IN & (1ULL << b)))) {
		return ESP_ERR_INVALID_ARG;
	}

    // Allocate space for the encoder
    encoder_h_t *encoder = calloc(1, sizeof(encoder_h_t));
    if (!encoder) {
		return ESP_ERR_NO_MEM;
    }

    encoder->A = a;
    encoder->B = b;
    encoder->state = R_START;
	encoder->counter = 0;

    //gpio configuration encoder a
    gpio_config_t io_conf;
    io_conf.intr_type = GPIO_PIN_INTR_DISABLE;
    io_conf.mode = GPIO_MODE_INPUT;
    io_conf.pin_bit_mask = (1ULL << a);
    io_conf.pull_down_en = 0;
    io_conf.pull_up_en = 1;

    gpio_config(&io_conf);

    //gpio configuration encoder b
    io_conf.pin_bit_mask = (1ULL << b);
    gpio_config(&io_conf);

    gpio_install_isr_service(0);

    gpio_set_pull_mode(a, GPIO_PULLUP_ONLY);
    gpio_set_intr_type(a, GPIO_INTR_ANYEDGE);
    gpio_isr_handler_add(a, encoder_isr, (void *)encoder);

    gpio_set_pull_mode(b, GPIO_PULLUP_ONLY);
    gpio_set_intr_type(b, GPIO_INTR_ANYEDGE);
    gpio_isr_handler_add(b, encoder_isr, (void *)encoder);


    *h = encoder;

    attached++;

    return ESP_OK;
}

esp_err_t encoder_unsetup(encoder_h_t *h) {
	portDISABLE_INTERRUPTS();

	if (attached == 0) {
		portENABLE_INTERRUPTS();
		return ESP_OK;
	}

	// Remove interrupts
	gpio_isr_handler_remove(h->A);
    gpio_isr_handler_remove(h->B);

	if (attached == 1) {
		if (task) {
			vTaskDelete(task);
			task = NULL;
		}

		if (queue) {
			vQueueDelete(queue);
			queue = NULL;
		}
	}

	attached--;

	free(h);

	portENABLE_INTERRUPTS();

	return ESP_OK;
}

esp_err_t encoder_read(encoder_h_t *h, int32_t *val) {
	portDISABLE_INTERRUPTS();
	*val = h->counter;
	portENABLE_INTERRUPTS();

	return ESP_OK;
}

esp_err_t encoder_write(encoder_h_t *h, int32_t val) {
	portDISABLE_INTERRUPTS();
	h->counter = val;
	portENABLE_INTERRUPTS();

	return ESP_OK;
}

esp_err_t encoder_register_callback(encoder_h_t *h, encoder_callback_t callback, int callback_id, uint8_t deferred) {
	portDISABLE_INTERRUPTS();

	h->callback = callback;
	h->callback_id = callback_id;
	h->deferred = deferred;

	if (deferred) {
		if (!queue) {
			queue = xQueueCreate(10, sizeof(encoder_deferred_data_t));
			if (!queue) {
				portENABLE_INTERRUPTS();
				return ESP_ERR_NO_MEM;
			}
		}

		if (!task) {
			BaseType_t xReturn;

			xReturn = xTaskCreatePinnedToCore(encoder_task, "encoder", CONFIG_ENCODER_THREAD_STACK_SIZE, NULL, CONFIG_ENCODER_THREAD_PRIORITY, &task, CONFIG_ENCODER_CPU);
			if (xReturn != pdPASS) {
				portENABLE_INTERRUPTS();
				return ESP_ERR_NO_MEM;
			}
		}
	}

	portENABLE_INTERRUPTS();

	return ESP_OK;
}
