#include <stdio.h>
#include "motors.h"
#include "driver/gpio.h"


#define GPIO_ALL_OUT 0b0000001100001110111111111111111111111111ULL
#define FRECUENCY 1

esp_err_t motor_install(int8_t a, int8_t b, mcpwm_unit_t pwm_unit, mcpwm_io_signals_t pwm_a, mcpwm_io_signals_t pwm_b, mcpwm_timer_t t, motor_h_t **m)
{

    if ((!GPIO_IS_VALID_GPIO(a)) && ((GPIO_ALL_OUT & (1ULL << a)))) {
		return ESP_ERR_INVALID_ARG;
	}

    if ((!GPIO_IS_VALID_GPIO(b)) && ((GPIO_ALL_OUT & (1ULL << b)))) {
		return ESP_ERR_INVALID_ARG;
	}

    motor_h_t *motor = calloc(1, sizeof(motor_h_t));
    
    if (!motor) {
		return ESP_ERR_NO_MEM;
    }

    motor->pin1 = a;
    motor->pin2 = b;
    motor->info = 0x00;
    motor->intSpeed = 0;
    motor->defaultDuty = 0;

    motor->mcpwm_unit = pwm_unit;
    motor->mcpwm_timer = t;
    motor->mcpwm_a = pwm_a;
    motor->mcpwm_b = pwm_b;

    mcpwm_gpio_init(pwm_unit, pwm_a, a);
    mcpwm_gpio_init(pwm_unit, pwm_b, b);

    *m = motor;

    return ESP_OK;
}

/**
 * @brief motor moves in forward direction, with duty cycle = duty %
 */
void motor_forward(mcpwm_unit_t mcpwm_num, mcpwm_timer_t timer_num , float duty_cycle)
{
    //printf("\t Motor speed: %.3f\n",duty_cycle);
    mcpwm_set_signal_low(mcpwm_num, timer_num, MCPWM_OPR_B);
    mcpwm_set_duty(mcpwm_num, timer_num, MCPWM_OPR_A, duty_cycle);
    mcpwm_set_duty_type(mcpwm_num, timer_num, MCPWM_OPR_A, MCPWM_DUTY_MODE_0); //call this each time, if operator was previously in low/high state
}

/**
 * @brief motor moves in backward direction, with duty cycle = duty %
 */
void motor_backward(mcpwm_unit_t mcpwm_num, mcpwm_timer_t timer_num , float duty_cycle)
{
    //printf("\t Motor speed: %.3f\n",duty_cycle);
    mcpwm_set_signal_low(mcpwm_num, timer_num, MCPWM_OPR_A);
    mcpwm_set_duty(mcpwm_num, timer_num, MCPWM_OPR_B, duty_cycle);
    mcpwm_set_duty_type(mcpwm_num, timer_num, MCPWM_OPR_B, MCPWM_DUTY_MODE_0);  //call this each time, if operator was previously in low/high state
}



/**
 * @brief set motor Speed with duty cycle = duty %
 */
esp_err_t motor_set_speed(motor_h_t *m, int intIn){
    
    if (m->info & 0x01){
        if ((intIn >= -100) && (intIn <= 100)) {
            m->intSpeed = intIn;
        }else {
            return ESP_ERR_INVALID_ARG;
        }

        if (intIn >= 0){
            motor_forward(m->mcpwm_unit, m->mcpwm_timer, (intIn*1.0));
        }else{
            motor_backward(m->mcpwm_unit, m->mcpwm_timer, (intIn*-1.0) );
        }
    }

    return ESP_OK;
}

/**
 * @brief motor stop
 */
esp_err_t motor_stop(motor_h_t *m)
{
    mcpwm_set_signal_low(m->mcpwm_unit, m->mcpwm_timer, MCPWM_OPR_A);
    mcpwm_set_signal_low(m->mcpwm_unit, m->mcpwm_timer, MCPWM_OPR_B);

    return ESP_OK;
}

/**
 * @brief motor start
 */
esp_err_t motor_start(motor_h_t *m)
{
    mcpwm_config_t pwm_config;
    pwm_config.frequency = 15000 * FRECUENCY;    //frequency = 500Hz,
    pwm_config.cmpr_a = 0;    //duty cycle of PWMxA = 0
    pwm_config.cmpr_b = 0;    //duty cycle of PWMxb = 0
    pwm_config.counter_mode = MCPWM_UP_COUNTER;
    pwm_config.duty_mode = MCPWM_DUTY_MODE_0;
    mcpwm_init(m->mcpwm_unit, m->mcpwm_timer, &pwm_config);    //Configure PWM0A & PWM0B with above settings

    m->info = m->info | 0x01;

    return ESP_OK;
}

